varying vec2 texcoord;

#include "/lib/sky_function.glsl"
float phaseg(float x, float g)
{
    float gg = g * g;
    return (gg * -0.25 + 0.25) * pow(-2.0 * (g * x) + (gg + 1.0), -1.5) / PI;
}

#if CLOUD_STYLE == 2
    #include "/lib/clouds/clouds_loop_a.glsl"
#else
    #include "/lib/clouds/clouds_bloop.glsl"
#endif

// https://software.intel.com/en-us/node/503873
vec3 RGB_YCoCg(vec3 c)
{
    // Y = R/4 + G/2 + B/4
    // Co = R/2 - B/2
    // Cg = -R/4 + G/2 - B/4
    return vec3(c.x / 4.0 + c.y / 2.0 + c.z / 4.0, c.x / 2.0 - c.z / 2.0, -c.x / 4.0 + c.y / 2.0 - c.z / 4.0);
}

void main()
{
    /* DRAWBUFFERS:8 */
    vec3 np3 = vec3(vec2(gbufferProjectionInverse[0].x, gbufferProjectionInverse[1].y) * (texcoord.xy * 2.0 - 1.0) +
                        gbufferProjectionInverse[3].xy,
                    1.0);
    np3 = vec3(np3.xy, gbufferProjectionInverse[3].z) / (gbufferProjectionInverse[2].w + gbufferProjectionInverse[3].w);
    np3 = mat3(gbufferModelViewInverse) * np3;
    np3 = normalize(np3);

#ifdef OVERWORLD
    vec4 skysample = texture(colortex8, texcoord.xy);
    vec3 color = skysample.rgb;
    vec4 cloud = vec4(0, 0, 0, 0);
    float alpha = skysample.a;

    #if CLOUD_STYLE == 1
    if (np3.y > 0.0)
    #endif

    {
        float noise = noise_standard(gl_FragCoord.xy);
    #ifndef ALT_CLOUDS
        #if CLOUD_STYLE == 2
        cloud = renderClouds(noise, np3, vec3(0.0), false, 1);

        #endif
        #if CLOUD_STYLE == 1
        cloud = renderClouds(noise, np3, color);
                // cloud = renderClouds(noise, np3, vec3(0), false);

        #endif
    #else
        cloud = vec4(0, 0, 0, 1);

    #endif
    }
    if (cloud.a < 0.2)
    {
        vec3 altcol = skyFromTex2((np3), colortex6).rgb;
        color = mix(altcol, color, cloud.a);
    }
    #if CLOUD_STYLE == 2
    color *= clamp(cloud.a, 0, 1);
    #endif
    color += cloud.rgb;
#else
    #ifdef END

    vec3 altcol = skyFromTex2((np3), colortex6).rgb;

    vec3 color = altcol;

    vec4 cloud = vec4(0, 0, 0, 0);
        #ifndef ALT_CLOUDS

    if (np3.y > 0.0)
    {
        float noise = noise_standard(gl_FragCoord.xy);
            #if CLOUD_STYLE == 2
        cloud = renderClouds(noise, np3, color, false, 1) * vec4(2, 0.5, 0.7, 1);

            #endif
            #if CLOUD_STYLE == 1
        cloud = renderClouds(noise, np3, color);
                    // cloud = renderClouds(noise, np3, vec3(0), false);

            #endif
    }

    color += cloud.rgb;
        #endif
    #else
    vec3 color = skyFromTex2((np3), colortex6).rgb;

    #endif
#endif
    color.rgb = Dither_Vlachos(color, 1024, gl_FragCoord.xy);

    gl_FragData[0].rgb = vec3(color);
}
