

#define waterfog
varying vec2 texcoord;

float GetBorderFogMixFactorCylinder(in vec3 eyePlayerPos, in float far, in float z0)
{

    float eyeDist = length(eyePlayerPos.xz);
    float borderFogFactor = smoothstep(far * 0.75, far, eyeDist);
    if (z0 >= 1)
        borderFogFactor = 1;
    return borderFogFactor;
}
float ld(const float dist)
{
    return (2.0 * near) / (far + near - dist * (far - near));
}
float ld_dh(const float dist)
{
    return (2.0 * dhNearPlane) / (dhFarPlane + dhNearPlane - dist * (dhFarPlane - dhNearPlane));
}

/////

/////

#include "/lib/ssaodh.glsl"

vec3 getDepthPoint(vec2 coord, float depth)
{
    vec4 pos;
    pos.xy = coord;
    pos.z = depth;
    pos.w = 1.0;
    pos.xyz = pos.xyz * 2.0 - 1.0; // convert from the 0-1 range to the -1 to +1 range
    pos = dhProjectionInverse * pos;
    pos.xyz /= pos.w;

    return pos.xyz;
}

vec3 constructNormal(float depthA, vec2 texcoords, sampler2D depthtex)
{
    const vec2 offsetB = vec2(0.0, 0.001);
    const vec2 offsetC = vec2(0.001, 0.0);

    float depthB = texture(depthtex, texcoords + offsetB).r;
    float depthC = texture(depthtex, texcoords + offsetC).r;

    vec3 A = getDepthPoint(texcoords, depthA);
    vec3 B = getDepthPoint(texcoords + offsetB, depthB);
    vec3 C = getDepthPoint(texcoords + offsetC, depthC);

    vec3 AB = normalize(B - A);
    vec3 AC = normalize(C - A);

    vec3 normal = -cross(AB, AC);
    // normal.z = -normal.z;

    return normalize(normal);
}
// https://gitlab.com/jeseibel/distant-horizons-core/-/blob/main/core/src/main/resources/shaders/flat_shaded.frag?ref_type=heads
// Property of Distant Horizons [mod]

const int noiseSteps = 16;
const float noiseIntensity = 16.0;
const int noiseDropoff = 1024;

float rand(float co)
{
    return fract(sin(co * (91.3458)) * 47453.5453);
}
float rand(vec2 co)
{
    return fract(sin(dot(co.xy, vec2(12.9898, 78.233))) * 43758.5453);
}
float rand(vec3 co)
{
    return rand(co.xy + rand(co.z));
}

vec3 quantize(const in vec3 val, const in int stepSize)
{
    return floor(val * stepSize) / stepSize;
}
float toLinear2(float sRGB)
{
    return sRGB * (sRGB * (sRGB * 0.305306011 + 0.682171111) + 0.012522878);
}

void applyNoise(inout vec3 fragColor, const in vec3 worldPos, const in float viewDist)
{
    float noiseAmplification = noiseIntensity * 0.01;
    float lum = (fragColor.r + fragColor.g + fragColor.b) / 3.0;
    noiseAmplification =
        (1.0 - pow(lum * 2.0 - 1.0, 2.0)) * noiseAmplification; // Lessen the effect on depending on how dark the object
                                                                // is, equasion for this is -(2x-1)^{2}+1

    // Random value for each position
    float randomValue = rand(quantize(worldPos, noiseSteps));
    randomValue = randomValue * 1.5 * noiseAmplification - noiseAmplification;
    randomValue = (randomValue);
    // Modifies the color
    // A value of 0 on the randomValue will result in the original color, while a value of 1 will result in a fully
    // bright color
    vec3 newCol = fragColor.rgb + (1.0 - fragColor.rgb) * randomValue;
    newCol = clamp(newCol, 0, 1);

    if (noiseDropoff != 0)
    {
        float distF = min(viewDist / noiseDropoff, 1.0);
        newCol = mix(newCol, fragColor.rgb, distF); // The further away it gets, the less noise gets applied
    }

    fragColor.rgb = newCol;
}

float linZ(float depth)
{
    return (2.0 * dhNearPlane) / (dhFarPlane + dhNearPlane - depth * (dhFarPlane - dhNearPlane));
}
float getRawDepth(vec2 uv)
{
    return texture(dhDepthTex0, uv).x;
}

vec3 viewSpacePosAtScreenUV(vec2 uv)
{
    vec3 viewSpaceRay = (dhProjectionInverse * vec4(uv * 2.0 - 1.0, 1.0, 1.0) * near).xyz;
    float rawDepth = getRawDepth(uv);
    return viewSpaceRay * (linZ(rawDepth));
}

vec3 viewNormalAtPixelPosition2(vec2 vpos)
{
    // screen uv from vpos
    vec2 uv = vpos * texelSize.xy;

    // current pixel's depth
    float c = getRawDepth(uv);

    // get current pixel's view space position
    vec3 viewSpacePos_c = viewSpacePosAtScreenUV(uv);

    // get view space position at 1 pixel offsets in each major direction
    vec3 viewSpacePos_l = viewSpacePosAtScreenUV(uv + vec2(-1.0, 0.0) * texelSize.xy);
    vec3 viewSpacePos_r = viewSpacePosAtScreenUV(uv + vec2(1.0, 0.0) * texelSize.xy);
    vec3 viewSpacePos_d = viewSpacePosAtScreenUV(uv + vec2(0.0, -1.0) * texelSize.xy);
    vec3 viewSpacePos_u = viewSpacePosAtScreenUV(uv + vec2(0.0, 1.0) * texelSize.xy);

    // get the difference between the current and each offset position
    vec3 l = viewSpacePos_c - viewSpacePos_l;
    vec3 r = viewSpacePos_r - viewSpacePos_c;
    vec3 d = viewSpacePos_c - viewSpacePos_d;
    vec3 u = viewSpacePos_u - viewSpacePos_c;

    // get depth values at 1 & 2 pixels offsets from current along the horizontal
    // axis
    vec4 H = vec4(getRawDepth(uv + vec2(-1.0, 0.0) * texelSize.xy), getRawDepth(uv + vec2(1.0, 0.0) * texelSize.xy),
                  getRawDepth(uv + vec2(-2.0, 0.0) * texelSize.xy), getRawDepth(uv + vec2(2.0, 0.0) * texelSize.xy));

    // get depth values at 1 & 2 pixels offsets from current along the vertical
    // axis
    vec4 V = vec4(getRawDepth(uv + vec2(0.0, -1.0) * texelSize.xy), getRawDepth(uv + vec2(0.0, 1.0) * texelSize.xy),
                  getRawDepth(uv + vec2(0.0, -2.0) * texelSize.xy), getRawDepth(uv + vec2(0.0, 2.0) * texelSize.xy));

    // current pixel's depth difference from slope of offset depth samples
    // differs from original article because we're using non-linear depth values
    // see article's comments
    vec2 he = abs((2 * H.xy - H.zw) - c);
    vec2 ve = abs((2 * V.xy - V.zw) - c);

    // pick horizontal and vertical diff with the smallest depth difference from
    // slopes
    vec3 hDeriv = he.x < he.y ? l : r;
    vec3 vDeriv = ve.x < ve.y ? d : u;

    // get view space normal from the cross product of the best derivatives
    vec3 viewNormal = normalize(cross(hDeriv, vDeriv));

    return viewNormal;
}
float scaleByDepth(float depth)
{
    float nearScale = 1.0;             // Scale factor near the camera
    float farScale = 0.1;              // Scale factor far from the camera
    float depthFactor = depth / 100.0; // Adjust this based on your scene depth
    return mix(nearScale, farScale, depthFactor);
}

float RT(vec3 dir, vec3 position, float noise)
{

    vec3 clipPosition = toClipSpace3DH(position);

    float rayLength = ((position.z + dir.z * sqrt(3.0) * dhFarPlane) > -sqrt(3.0) * dhNearPlane) ? (-sqrt(3.0) * dhNearPlane - position.z) / dir.z : sqrt(3.0) * dhFarPlane;

    vec3 end = toClipSpace3DH(position + dir * rayLength);
    vec3 direction = end - clipPosition;
    float len = max(abs(direction.x) / texelSize.x, abs(direction.y) / texelSize.y);

    vec3 stepv = direction / len;

    float scaledNoise = 24.0 * noise + 1.0;

    vec3 scaledStep = stepv * scaledNoise;

    vec3 spos = clipPosition + scaledStep;

    spos.xy += offsets[framemod8] * texelSize * 0.5;

    float sp = linZ(texelFetch(dhDepthTex0, ivec2(spos.xy / texelSize), 0).r);

    float currZ = linZ(spos.z);
    float out1 = 1.0 - float((sp < currZ && distance(sp, currZ) / currZ < noise * 0.2));

    return out1;
}

vec2 R2_samples(int n)
{
    vec2 alpha = vec2(0.75487765, 0.56984026);
    return fract(alpha * n);
}
vec3 GenerateUnitVector(vec2 xy)
{
    const float tau = radians(360.0);

    xy.x *= tau;
    xy.y = xy.y * 2.0 - 1.0;
    return vec3(vec2(cos(xy.x), sin(xy.x)) * sqrt(1.0 - xy.y * xy.y), xy.y);
}

vec3 GenerateCosineVector(vec3 vector, vec2 xy)
{

    return normalize(vector + GenerateUnitVector(xy));
}

float rtao(float skyLight, vec3 normal, vec3 fragpos)
{

    int samples = 4;
    float occlusion = 0.0;
    for (int i = 0; i < samples; i++)
    {
        int seed = frameCounter * int(samples) + i;
        vec2 ij = fract(R2_samples(seed) + blueNoise4(gl_FragCoord.xy).rg);
        vec3 L = GenerateCosineVector(normal, ij);

        occlusion += RT(mat3(gbufferModelView) * L, fragpos.xyz, blueNoise(gl_FragCoord.xy));
    }

    return pow(occlusion / float(samples), 2.0);
}

void main()
{

    /* DRAWBUFFERS:0 */

    float z = texture(depthtex0, texcoord).x;

    float z0DH = texture(dhDepthTex0, texcoord).x;
    vec3 DH = texture(colortex10, texcoord).rgb;
    vec3 sky = texture(colortex8, texcoord).rgb;

    if (z0DH >= 1.0)
    {
        DH = sky;
    }
    float noise = noise_standard(gl_FragCoord.xy) * 0.001;

    vec3 fragposDH = toScreenSpaceDH(vec3(texcoord, z0DH));
    vec3 fragpos = toScreenSpace(vec3(texcoord, z));

    vec3 p3 = mat3(gbufferModelViewInverse) * fragposDH + gbufferModelViewInverse[3].xyz;
    vec3 p3_org = mat3(gbufferModelViewInverse) * fragpos + gbufferModelViewInverse[3].xyz;
    vec3 worldPos = p3 + cameraPosition;
    float viewDist = length(fragposDH);

    // applyNoise(DH, worldPos + noise, viewDist);

    float mixfactor = GetBorderFogMixFactorCylinder(p3_org, far, z);
    vec3 color = texture(colortex0, texcoord.xy).rgb;

    // vec3 viewNormal = cross(dFdx(fragposDH), dFdy(fragposDH));
    vec3 viewNormal = viewNormalAtPixelPosition2(gl_FragCoord.xy);
    vec3 worldnormal = mat3(gbufferModelViewInverse) * viewNormal;

    if (mixfactor > 0 && z0DH < 1.0)
    {

        DH *= rtao(0, worldnormal, fragposDH);
    }

    color.rgb = mix(color.rgb, DH, mixfactor);

    gl_FragData[0].rgb = vec3(color);
}
