

varying vec2 texcoord;

vec3 median(sampler2D tex1, ivec2 coords)
{
    vec3 v[9];

    // Fill the array 'v' with the color values from the 3x3 pixel window
    for (int dX = -1; dX <= 1; ++dX)
    {
        for (int dY = -1; dY <= 1; ++dY)
        {
            ivec2 offset = ivec2(dX, dY);
            v[(dX + 1) * 3 + (dY + 1)] = texelFetch(tex1, coords + offset, 0).rgb;
        }
    }

    // Selection sort to find the 5th smallest element
    for (int i = 0; i < 5; ++i)
    {
        int minIndex = i;
        for (int j = i + 1; j < 9; ++j)
        {
            if (v[j].r + v[j].g + v[j].b < v[minIndex].r + v[minIndex].g + v[minIndex].b)
            {
                minIndex = j;
            }
        }

        // Swap to bring the current smallest to the i-th position
        vec3 temp = v[i];
        v[i] = v[minIndex];
        v[minIndex] = temp;
    }

    // Return the median value, which is the 5th smallest element
    return v[4];
}

vec3 tonemap(vec3 col)
{
    // return col;
    return col / (1 + luma(col));
}
vec3 invTonemap(vec3 col)
{
    // return col;

    return col / (1 - luma(col));
}

float MinDepth3x3(vec2 uv)
{
    ivec2 fragCoord = ivec2(gl_FragCoord.xy);
    float depth = texelFetch(depthtex0, fragCoord, 0).x;

    for (int x = -1; x <= 1; ++x)
    {
        for (int y = -1; y <= 1; ++y)
        {
            if (x != 0 || y != 0)
            {
                depth = min(depth, texelFetch(depthtex0, fragCoord + ivec2(x, y), 0).x);
            }
        }
    }

    return depth;
}
vec3 untonemap(vec3 inp)
{
    return invTonemap(inp);
}
vec4 textureYC(sampler2D tex, vec2 uv)
{
    vec4 c = texture2D(tex, uv);
    //  c.rgb = median(tex, ivec2(uv * viewResolution));

    return vec4(tonemap(c.rgb), c.a);
}
vec4 texelFetchYC(sampler2D tex, ivec2 uv, float numb)
{
    vec4 c = texelFetch(tex, ivec2(uv), 0);
    //    c.rgb = median(tex, uv);

    return vec4(tonemap(c.rgb), c.a);
}
#ifdef DISTANT_HORIZONS

float MinDepth3x3DH(vec2 uv)
{
    ivec2 fragCoord = ivec2(gl_FragCoord.xy);
    float depth = texelFetch(dhDepthTex0, fragCoord, 0).x;

    for (int x = -1; x <= 1; ++x)
    {
        for (int y = -1; y <= 1; ++y)
        {
            if (x != 0 || y != 0)
            {
                depth = min(depth, texelFetch(dhDepthTex0, fragCoord + ivec2(x, y), 0).x);
            }
        }
    }

    return depth;
}

vec3 ReprojectDH(vec3 position)
{
    bool isHand = position.z < 0.6;
    if (position.z < 1.0 && !isHand)
    {
        position = ViewSpaceFromScreenSpace(position, dhProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position += (cameraPosition - previousCameraPosition);
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, dhPreviousProjection);
    }
    else
    {
        position = ViewSpaceFromScreenSpace(position, dhProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, dhPreviousProjection);
    }

    return position;
}
#endif
vec3 Reproject(vec3 position)
{
    bool isHand = position.z < 0.6;
    if (position.z < 1.0 && !isHand)
    {
        position = ViewSpaceFromScreenSpace(position, gbufferProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position += (cameraPosition - previousCameraPosition);
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, gbufferPreviousProjection);
    }
    else
    {
        position = ViewSpaceFromScreenSpace(position, gbufferProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, gbufferPreviousProjection);
    }

    return position;
}

// Fetch history pixel
vec4 fetchHistoryPx(ivec2 pix)
{
    return vec4(texelFetchYC(colortex9, pix, 0).xyz, 1.0);
}

// Cubic Hermite interpolation
vec3 CubicHermite(vec3 A, vec3 B, vec3 C, vec3 D, float t)
{
    float t2 = t * t;
    float t3 = t2 * t;
    vec3 a = -0.5 * A + 1.5 * B - 1.5 * C + 0.5 * D;
    vec3 b = A - 2.5 * B + 2.0 * C - 0.5 * D;
    vec3 c = -0.5 * A + 0.5 * C;
    vec3 d = B;
    return a * t3 + b * t2 + c * t + d;
}

vec3 BicubicHermiteTextureSample(vec2 P)
{
    vec2 pixel = P * viewResolution + 0.5;
    vec2 frac = fract(pixel);
    ivec2 ipixel = ivec2(pixel) - 1;

    vec3 C00 = fetchHistoryPx(ipixel + ivec2(-1, -1)).rgb;
    vec3 C10 = fetchHistoryPx(ipixel + ivec2(0, -1)).rgb;
    vec3 C20 = fetchHistoryPx(ipixel + ivec2(1, -1)).rgb;
    vec3 C30 = fetchHistoryPx(ipixel + ivec2(2, -1)).rgb;

    vec3 C01 = fetchHistoryPx(ipixel + ivec2(-1, 0)).rgb;
    vec3 C11 = fetchHistoryPx(ipixel + ivec2(0, 0)).rgb;
    vec3 C21 = fetchHistoryPx(ipixel + ivec2(1, 0)).rgb;
    vec3 C31 = fetchHistoryPx(ipixel + ivec2(2, 0)).rgb;

    vec3 C02 = fetchHistoryPx(ipixel + ivec2(-1, 1)).rgb;
    vec3 C12 = fetchHistoryPx(ipixel + ivec2(0, 1)).rgb;
    vec3 C22 = fetchHistoryPx(ipixel + ivec2(1, 1)).rgb;
    vec3 C32 = fetchHistoryPx(ipixel + ivec2(2, 1)).rgb;

    vec3 C03 = fetchHistoryPx(ipixel + ivec2(-1, 2)).rgb;
    vec3 C13 = fetchHistoryPx(ipixel + ivec2(0, 2)).rgb;
    vec3 C23 = fetchHistoryPx(ipixel + ivec2(1, 2)).rgb;
    vec3 C33 = fetchHistoryPx(ipixel + ivec2(2, 2)).rgb;

    vec3 CP0X = CubicHermite(C00, C10, C20, C30, frac.x);
    vec3 CP1X = CubicHermite(C01, C11, C21, C31, frac.x);
    vec3 CP2X = CubicHermite(C02, C12, C22, C32, frac.x);
    vec3 CP3X = CubicHermite(C03, C13, C23, C33, frac.x);

    return CubicHermite(CP0X, CP1X, CP2X, CP3X, frac.y);
}

vec3 clip_aabb(vec3 min_color, vec3 max_color, vec3 history_color)
{
    vec3 p_clip = 0.5 * (max_color + min_color);
    vec3 e_clip = 0.5 * (max_color - min_color);

    vec3 v_clip = history_color - p_clip;
    vec3 v_unit = v_clip / e_clip;
    vec3 a_unit = abs(v_unit);
    float ma_unit = max(a_unit.x, max(a_unit.y, a_unit.z));
    bool history_clipped = ma_unit > 1.0;
    vec3 history = history_clipped ? p_clip + v_clip / ma_unit : history_color;

    return history;
}
vec3 clip_aabb_simple(vec3 min_color, vec3 max_color, vec3 history_color)
{
    float tolerance = 0.0;

    // Expand the AABB by the tolerance value
    vec3 min_color_tolerant = min_color - tolerance;
    vec3 max_color_tolerant = max_color + tolerance;

    // Clamp the history color to the extended bounds of the AABB
    vec3 clipped_color = clamp(history_color, min_color_tolerant, max_color_tolerant);

    // Return the clipped color
    return clipped_color;
}

vec3 processTexel(vec3 scene2, vec3 scene, vec3 albedo, ivec2 texel, ivec2 du, ivec2 dv)
{
    // Fetch and decode texel, returning the processed color
    vec3 a, b;

    // Fetch the texel, decode two vec3s using decodeTwoVec3
    decodeTwoVec3(texelFetch(colortex12, texel, 0), a, b);

    // Calculate the output color based on scene and metallic value
    vec3 out1 = a + b * scene;

    out1 *= albedo;

    return tonemap((out1 - scene2));
}
vec3 sampleBlur(
    sampler2D tex,
    vec2 res,
    vec2 pos,
    float sceneDepth,
    vec3 sceneNormal,
    float directions,
    float quality,
    float size)
{
    const float TAU = PI * 2.0;

    vec2 radius = size / res.xy;
    vec2 uv = pos / res.xy;

    vec3 color = texture(tex, uv).rgb;

    float directionStep = TAU / directions;
    float qualityStep = 1.0 / quality;

    float importance = 1.0;

    for (float arc = 0.0; arc < TAU; arc += directionStep)
        for (float q = 1.0; q > 0.0; q -= qualityStep)
        {
            vec2 offset = vec2(cos(arc), sin(arc)) * q * radius * res.y;
            vec2 cuv = uv + offset;

            // Avoid sampling outside the edges
            if (
                cuv.x < 0.0 ||
                cuv.y < 0.0 ||
                cuv.x > 1.0 ||
                cuv.y > 1.0)
            {
                continue;
            }

            vec4 fog = texture(tex, cuv);
            vec3 sampleNormal = decodeNormal(texture(colortex2, cuv).rg);
            float spatialImportance = 1.0 + (1.0 - q); // linear falloff
            float depthImportance = clamp(1.0 / abs(sceneDepth - fog.w), 0.0, 1.0);

            float normalDot = max(dot(sceneNormal, sampleNormal), 0.0);
            float normalImportance = normalDot; // using the dot product for normal importance

            float sampleImportance = spatialImportance * depthImportance * normalImportance;
            importance += sampleImportance;
            color += texture(tex, cuv).rgb * sampleImportance;
        }

    return color / importance;
}

void main()
{

    /* RENDERTARGETS: 9,0,1 */

    vec3 albedo = texture(colortex1, texcoord).rgb;

    vec2 texSize = textureSize(colortex12, 0); // Get the size of the base mipmap level (LOD 0)

    vec4 materials = texture(colortex3, texcoord);

    vec2 decoded2 = decodeVec2(materials.x);
    float smoothness = decoded2.y;

    int metalidx = int((decoded2.x * 255));
    float historyWeight = mix(0.99, 0.90, smoothness);

    float offcenter_rej = 0.1;

    vec3 color = vec3(0.0);
    vec2 uv = texcoord;
    float z = texture(depthtex0, texcoord).x;

    ivec2 texel = ivec2(uv * texSize);
    vec3 position = vec3(uv, MinDepth3x3(texel));
    vec3 reprojectedPosition = Reproject(position);

    bool hand = texture(depthtex2, texcoord).x > z;
    vec3 scene = texture(colortex0, texcoord).rgb;
    vec3 scene2 = scene;
    if (hand)
        historyWeight = 0.0;

    if (z >= 1)
    {
        gl_FragData[1].rgb = scene;
        return;
    }

    if (metalidx < 229)
    {
        albedo = vec3(1.0);
    }
    else
    {
        scene2 = vec3(0);
    }
    //   albedo = vec3(1.0);
    vec3 test1;
    vec3 test2;

    decodeTwoVec3(texelFetch(colortex12, ivec2(texel + offsets[framemod8]), 0), test1, test2);

    vec3 current = tonemap((((test1 + test2 * scene) * albedo) - scene2) + 0.0);


    //  current = test;
    // vec3 history = textureYC(colortex9, reprojectedPosition.xy).rgb;
    vec3 history = (BicubicHermiteTextureSample(reprojectedPosition.xy));

    vec2 pixelCenterDist = 1.0 - abs(2.0 * fract(reprojectedPosition.xy * viewResolution) - 1.0);
    historyWeight *= sqrt(pixelCenterDist.x * pixelCenterDist.y) * offcenter_rej + (1.0 - offcenter_rej);
    //  historyWeight = 0;
    ivec2 du = ivec2(1.0, 0.0);
    ivec2 dv = ivec2(0.0, 1.0);

    vec3 ctl = processTexel(scene2, scene, albedo, texel - dv - du, du, dv);
    vec3 ctc = processTexel(scene2, scene, albedo, texel - dv, du, dv);
    vec3 ctr = processTexel(scene2, scene, albedo, texel - dv + du, du, dv);
    vec3 cml = processTexel(scene2, scene, albedo, texel - du, du, dv);
    vec3 cmc = current;
    vec3 cmr = processTexel(scene2, scene, albedo, texel + du, du, dv);
    vec3 cbl = processTexel(scene2, scene, albedo, texel + dv - du, du, dv);
    vec3 cbc = processTexel(scene2, scene, albedo, texel + dv, du, dv);
    vec3 cbr = processTexel(scene2, scene, albedo, texel + dv + du, du, dv);

    vec3 minRounded = min(ctl, min(ctc, min(ctr, min(cml, min(cmc, min(cmr, min(cbl, min(cbc, cbr))))))));
    vec3 maxRounded = max(ctl, max(ctc, max(ctr, max(cml, max(cmc, max(cmr, max(cbl, max(cbc, cbr))))))));

    history = clip_aabb_simple(minRounded, maxRounded, history).rgb;

    color = mix(current, history, historyWeight);
    color = untonemap(color);

    vec3 outc;
    outc = color;

    color += scene2;

    gl_FragData[0].rgb = outc;
    gl_FragData[1].rgb = color;
    gl_FragData[2] = vec4(0.0);

#if DEBUG_MODE == 1
    gl_FragData[1].rgb = scene;
#elif DEBUG_MODE == 2
    gl_FragData[1].rgb = color;
#elif DEBUG_MODE == 3
    gl_FragData[1].rgb = scene;
#elif DEBUG_MODE == 4
    gl_FragData[1].rgb = scene;
#elif DEBUG_MODE == 5
    gl_FragData[1].rgb = scene;
#endif
}
