

varying vec2 texcoord;

#include "/lib/sspt_lib.glsl"

vec2 texSize = 1.0 / (viewResolution / textureSize(colortex13, 0));

vec3 rgb2xyz(vec3 rgb)
{
    float r = rgb.r / 255.0;
    float g = rgb.g / 255.0;
    float b = rgb.b / 255.0;

    r = (r > 0.04045) ? pow((r + 0.055) / 1.055, 2.4) : r / 12.92;
    g = (g > 0.04045) ? pow((g + 0.055) / 1.055, 2.4) : g / 12.92;
    b = (b > 0.04045) ? pow((b + 0.055) / 1.055, 2.4) : b / 12.92;

    r *= 100.0;
    g *= 100.0;
    b *= 100.0;

    float x = r * 0.4124 + g * 0.3576 + b * 0.1805;
    float y = r * 0.2126 + g * 0.7152 + b * 0.0722;
    float z = r * 0.0193 + g * 0.1192 + b * 0.9505;

    return vec3(x, y, z);
}

vec3 xyz2lab(vec3 xyz)
{
    float x = xyz.x / 95.047;
    float y = xyz.y / 100.0;
    float z = xyz.z / 108.883;

    x = (x > 0.008856) ? pow(x, 1.0 / 3.0) : (7.787 * x) + (16.0 / 116.0);
    y = (y > 0.008856) ? pow(y, 1.0 / 3.0) : (7.787 * y) + (16.0 / 116.0);
    z = (z > 0.008856) ? pow(z, 1.0 / 3.0) : (7.787 * z) + (16.0 / 116.0);

    float l = (116.0 * y) - 16.0;
    float a = 500.0 * (x - y);
    float b = 200.0 * (y - z);

    return vec3(l, a, b);
}

vec3 rgb2lab(vec3 rgb)
{
    return xyz2lab(rgb2xyz(rgb));
}
vec3 lab2xyz(vec3 lab)
{
    float y = (lab.x + 16.0) / 116.0;
    float x = lab.y / 500.0 + y;
    float z = y - lab.z / 200.0;

    x = (pow(x, 3.0) > 0.008856) ? pow(x, 3.0) : (x - 16.0 / 116.0) / 7.787;
    y = (pow(y, 3.0) > 0.008856) ? pow(y, 3.0) : (y - 16.0 / 116.0) / 7.787;
    z = (pow(z, 3.0) > 0.008856) ? pow(z, 3.0) : (z - 16.0 / 116.0) / 7.787;

    x *= 95.047;
    y *= 100.0;
    z *= 108.883;

    return vec3(x, y, z);
}

vec3 xyz2rgb(vec3 xyz)
{
    float x = xyz.x / 100.0;
    float y = xyz.y / 100.0;
    float z = xyz.z / 100.0;

    float r = x * 3.2406 + y * -1.5372 + z * -0.4986;
    float g = x * -0.9689 + y * 1.8758 + z * 0.0415;
    float b = x * 0.0557 + y * -0.2040 + z * 1.0570;

    r = (r > 0.0031308) ? 1.055 * pow(r, 1.0 / 2.4) - 0.055 : 12.92 * r;
    g = (g > 0.0031308) ? 1.055 * pow(g, 1.0 / 2.4) - 0.055 : 12.92 * g;
    b = (b > 0.0031308) ? 1.055 * pow(b, 1.0 / 2.4) - 0.055 : 12.92 * b;

    return vec3(r, g, b) * 255.0;
}

vec3 lab2rgb(vec3 lab)
{
    return xyz2rgb(lab2xyz(lab));
}
vec3 tonemap(vec3 col)
{
    return col / (1 + luma(col));
}
vec3 invTonemap(vec3 col)
{
    return col / (1 - luma(col));
}
//  https://software.intel.com/en-us/node/503873
vec3 RGB_YCoCg(vec3 c)
{
    // Y = R/4 + G/2 + B/4
    // Co = R/2 - B/2
    // Cg = -R/4 + G/2 - B/4

    return vec3(c.x / 4.0 + c.y / 2.0 + c.z / 4.0, c.x / 2.0 - c.z / 2.0, -c.x / 4.0 + c.y / 2.0 - c.z / 4.0);
}

// https://software.intel.com/en-us/node/503873
vec3 YCoCg_RGB(vec3 c)
{
    // R = Y + Co - Cg
    // G = Y + Cg
    // B = Y - Co - Cg

    return cout(vec3(c.x + c.y - c.z, c.x + c.z, c.x - c.y - c.z));
}

vec4 textureYC(sampler2D tex, vec2 uv)
{
    vec4 c = texture2D(tex, uv);
    return vec4((c.rgb), c.a);
}
vec4 texelFetchYC(sampler2D tex, ivec2 uv, float numb)
{
    vec4 c = texelFetch(tex, ivec2(uv), 0);
    return vec4((c.rgb), c.a);
}

float MinDepth3x3(vec2 uv)
{
    ivec2 texel = ivec2(texcoord * viewResolution * texSize);

    float depth = texelFetch(depthtex0, texel, 0).x;

    for (int x = -1; x <= 1; ++x)
    {
        for (int y = -1; y <= 1; ++y)
        {
            if (x != 0 || y != 0)
            {
                depth = min(depth, texelFetch(depthtex0, texel + ivec2(x, y), 0).x);
            }
        }
    }

    return depth;
}

vec3 Reproject(vec3 position)
{
    bool isHand = position.z < 0.6;
    if (position.z < 1.0 && !isHand)
    {
        position = ViewSpaceFromScreenSpace(position, gbufferProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position += (cameraPosition - previousCameraPosition);
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, gbufferPreviousProjection);
    }
    else
    {
        position = ViewSpaceFromScreenSpace(position, gbufferProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, gbufferPreviousProjection);
    }

    return position;
}

vec3 clip_aabb(vec3 min_color, vec3 max_color, vec3 history_color)
{
    vec3 p_clip = 0.5 * (max_color + min_color);
    vec3 e_clip = 0.5 * (max_color - min_color);

    vec3 v_clip = history_color - p_clip;
    vec3 v_unit = v_clip / e_clip;
    vec3 a_unit = abs(v_unit);
    float ma_unit = max(a_unit.x, max(a_unit.y, a_unit.z));
    bool history_clipped = ma_unit > 1.0;
    vec3 history = history_clipped ? p_clip + v_clip / ma_unit : history_color;

    return history;
}
vec3 clip_aabb_simple(vec3 min_color, vec3 max_color, vec3 history_color)
{
    float tolerance = 0.0;

    // Expand the AABB by the tolerance value
    vec3 min_color_tolerant = min_color - tolerance;
    vec3 max_color_tolerant = max_color + tolerance;

    // Clamp the history color to the extended bounds of the AABB
    vec3 clipped_color = clamp(history_color, min_color_tolerant, max_color_tolerant);

    // Return the clipped color
    return clipped_color;
}
vec4 smoothfilter(in sampler2D tex, in vec2 uv)
{
    vec2 textureResolution = vec2(viewWidth, viewHeight);
    uv = uv * textureResolution + 0.5;
    vec2 iuv = floor(uv);
    vec2 fuv = fract(uv);
    uv = iuv + fuv * fuv * fuv * (fuv * (fuv * 6.0 - 15.0) + 10.0);
    uv = (uv - 0.5) / textureResolution;
    return textureYC(tex, uv);
}
//////////////////////////////////
const ivec2 kernelO_3x3[9] = ivec2[9](ivec2(-1, -1), ivec2(0, -1), ivec2(1, -1), ivec2(-1, 0), ivec2(0, 0), ivec2(1, 0),
                                      ivec2(-1, 1), ivec2(0, 1), ivec2(1, 1));
vec3 screenToViewSpace(vec3 screenpos, mat4 projInv, const bool taaAware)
{
    screenpos = screenpos * 2.0 - 1.0;

    vec3 viewpos = vec3(vec2(projInv[0].x, projInv[1].y) * screenpos.xy + projInv[3].xy, projInv[3].z);
    viewpos /= projInv[2].w * screenpos.z + projInv[3].w;

    return viewpos;
}
vec3 screenToViewSpace(vec3 screenpos, mat4 projInv)
{
    return screenToViewSpace(screenpos, projInv, true);
}

vec3 screenToViewSpace(vec3 screenpos)
{
    return screenToViewSpace(screenpos, gbufferProjectionInverse);
}
vec3 screenToViewSpace(vec3 screenpos, const bool taaAware)
{
    return screenToViewSpace(screenpos, gbufferProjectionInverse, taaAware);
}
float sqr(float x)
{
    return x * x;
}

const float rlog2 = 1.0 / log(2.0);
#define expf(x) exp2((x) * rlog2)

float computeVariance(sampler2D tex, ivec2 pos)
{
    float sum_msqr = 0.0;
    float sum_mean = 0.0;

    for (int i = 0; i < 9; i++)
    {
        ivec2 deltaPos = kernelO_3x3[i];

        vec3 col = texelFetch2D(tex, pos + deltaPos, 0).rgb;
        float lum = luma(col);

        sum_msqr += sqr(lum);
        sum_mean += lum;
    }
    sum_msqr /= 9.0;
    sum_mean /= 9.0;

    return abs(sum_msqr - sqr(sum_mean)) * 1 / (max(sum_mean, 1e-25));
}

#ifdef power

float Pow2(float x)
{
    return x * x;
}
vec2 Pow2(vec2 x)
{
    return x * x;
}
vec3 Pow2(vec3 x)
{
    return x * x;
}
vec4 Pow2(vec4 x)
{
    return x * x;
}
float Pow3(float x)
{
    return x * x * x;
}
float Pow4(float x)
{
    x *= x;
    return x * x;
}
vec2 Pow4(vec2 x)
{
    x *= x;
    return x * x;
}
vec3 Pow4(vec3 x)
{
    x *= x;
    return x * x;
}
float Pow5(float x)
{
    float x2 = x * x;
    return x2 * x2 * x;
}
float Pow6(float x)
{
    x *= x;
    return x * x * x;
}
float Pow8(float x)
{
    x *= x;
    x *= x;
    return x * x;
}
float Pow12(float x)
{
    x *= x;
    x *= x;
    return x * x * x;
}
float Pow16(float x)
{
    x *= x;
    x *= x;
    x *= x;
    return x * x;
}

#endif

vec3 atrous3(vec2 coord)
{

    int size = 1;
    ivec2 pos = ivec2(floor(coord * vec2(viewWidth, viewHeight)));
    vec3 colorCenter = texelFetch(colortex13, pos, 0).rgb;
    float variance = computeVariance(colortex13, ivec2(floor(coord * vec2(viewWidth, viewHeight))));
    if (variance == 0)
        return colorCenter;
    vec3 normals = decodeNormal(texelFetch(colortex2, pos, 0).rg);
    float z_org = texture(depthtex0, texcoord).x;

    float c_depth = z_org * far;
    vec3 origNormal = normals;

    vec3 totalColor = colorCenter;

    float totalWeight = 1.0;

    float var2 = 1 / (0.5 + variance * 4);

    // #define HQ

    const int r = 1;

    for (int i = 0; i < 9; i++)
    {
        ivec2 delta = kernelO_3x3[i] * size;

        if (delta.x == 0 && delta.y == 0)
            continue;

        ivec2 d_pos = pos + delta;

        float cu_depth = (texelFetch(depthtex0, d_pos, 0).x) * far;

        vec3 normal = decodeNormal(texelFetch(colortex2, d_pos, 0).rg);

        vec3 color = texelFetch(colortex13, d_pos, 0).rgb;
        if (distance(luma(color), luma(colorCenter)) < 0.0)
            return totalColor;
        float variance2 = computeVariance(colortex8, d_pos);

        float d_weight = abs(cu_depth - c_depth);
        float depthWeight = expf(-d_weight);
        if (depthWeight < 1e-5)
            continue;

        float normalWeight = pow(clamp(dot(normal, origNormal), 0, 1), 32);

        float weight = normalWeight;

        weight *= exp(-d_weight - var2);

        totalColor += color.rgb * weight;

        totalWeight += weight;
    }

    totalColor *= 1 / (max(totalWeight, 1e-25));

    return totalColor;
}
////////////////

vec2 tapLocation(int sampleNumber, float spinAngle, int nb, float nbRot)
{
    float alpha = float(sampleNumber + spinAngle / 6.28) / nb;
    float angle = alpha * (nbRot * 6.28) + spinAngle;

    float ssR = alpha * alpha;
    float sin_v, cos_v;

    sin_v = sin(angle);
    cos_v = cos(angle);

    return vec2(cos_v, sin_v) * ssR;
}

vec3 adjustSaturation(vec3 color, float userSaturation)
{
    float saturation = (userSaturation + 100.0) / 100.0;        // Normalize to 0 - 2
    float luminance = dot(color, vec3(0.2126, 0.7152, 0.0722)); // Luminance in linear space
    return mix(vec3(luminance), color, saturation);             // Adjust saturation linearly
}
float ld(const float dist)
{
    return (2.0 * near) / (far + near - dist * (far - near));
}

vec2 moment(ivec2 pos, vec2 momentIN)
{

    float weightSum = 1.0;
    vec2 moment = momentIN;
    float depth = texelFetch(depthtex0, pos, 0).x;
    vec3 normal = (decodeNormal(texelFetch(colortex2, pos, 0).rg));

    for (int i = 0; i < 9; i++)
    {
        ivec2 deltaPos = kernelO_3x3[i] * 2;

        // ⬇️ Sample current point data with current uv
        ivec2 p = pos + deltaPos;
        vec4 curColor = texelFetch(colortex13, ivec2(p), 0);
        float curDepth = texelFetch(depthtex0, p, 0).x;
        vec3 curNormal = (decodeNormal(texelFetch(colortex2, p, 0).rg));

        float l = luma(curColor.rgb);

        float weightDepth = abs(curDepth - depth) / (depth * length(vec2(deltaPos)) + 1.0e-2);
        float weightNormal = pow(max(0, dot(curNormal, normal)), 32.0);
        float w = exp(-weightDepth) * weightNormal;

        weightSum += w;

        moment += vec2(l, l * l) * w;
    }

    moment /= weightSum;

    return moment;
}

void main()
{

    /* RENDERTARGETS: 13,15 */

    vec3 color = vec3(0.0);
    vec2 uv = texcoord;
    float z = texture(depthtex0, texcoord).x;
    bool hand = texture(depthtex2, texcoord).x > z;
    vec4 normalTex = texture(colortex2, texcoord);
    vec3 albedo = texture(colortex0, texcoord).rgb;

    vec2 lightmap = decodeVec2(normalTex.b);
    if (z >= 1)
    {
        gl_FragData[0].rgb = atmosphereUp * 2.0;

        return;
    }
    if (hand)
    {
        gl_FragData[0].rgb = atmosphereUp * 2.0 * lightmap.y;

        return;
    }

    ////

    vec3 normal = (decodeNormal(normalTex.rg));
    vec4 materials = texture(colortex3, texcoord);

    float emissive = materials.y;

#ifdef TAA
    vec2 tempOffset = offsets[framemod8] * texelSize * 0.5 * texSize;
    vec3 fragpos = toScreenSpace(vec3(texcoord - tempOffset, z));
#else
    vec3 fragpos = toScreenSpace(vec3(texcoord, z));
#endif

    vec3 sspt =
        diffuseFunc(lightmap.y, normal, fragpos, (emissive * albedo * 32.0) + (albedo)) + albedo * emissive * 8;
    float lum = luma(sspt);
    vec3 diff = sspt - lum;
    sspt = sspt + diff * (-lum * 0.1 + 0.1);

    ////
    vec3 closestToCamera = vec3(texcoord, z);

    ivec2 texel = ivec2(uv * viewResolution * texSize);

    vec3 fragposition = fragpos;
    fragposition = mat3(gbufferModelViewInverse) * fragposition + gbufferModelViewInverse[3].xyz + (cameraPosition - previousCameraPosition);
    vec3 previousPosition = mat3(gbufferPreviousModelView) * fragposition + gbufferPreviousModelView[3].xyz;
    previousPosition = toClipSpace3Prev(previousPosition);
    vec2 velocity = previousPosition.xy - closestToCamera.xy;
    previousPosition.xy = texcoord + velocity;

    vec3 current = (sspt);

    vec3 history = textureYC(colortex13, previousPosition.xy).rgb;
    float historyA = textureYC(colortex13, previousPosition.xy).a;
    ///
    bool offscreen = clamp(previousPosition, 0, 1) != previousPosition;
    vec3 cameraMovement = mat3(gbufferModelView) * (cameraPosition - previousCameraPosition);
    vec4 currentGData = vec4(normal, 0);
    currentGData.a = length(fragpos);

    vec4 historyGData_0 = texture(colortex15, previousPosition.xy);
    vec4 historyGData = vec4(unpack2(historyGData_0.z), historyGData_0.w);

    vec2 moment = moment(ivec2((previousPosition.xy * viewResolution)), historyGData_0.rg);
    float distanceDelta = distance(historyGData.a * far, currentGData.a) - abs(cameraMovement.z);
    float depthRejection = (offscreen) ? 0.0 : exp(-max(distanceDelta - 0.2, 0.0) * 3.0);
    float normalWeight = sqr(clamp(dot(historyGData.rgb, currentGData.rgb), 0, 1));

    float accumulationR0 = mix(0.1, 0.02, sqrt(historyA));

    float varianceSpatial = clamp((max(0.0, moment.y - moment.x * moment.x)), 0, 1);
    varianceSpatial = accumulationR0 + varianceSpatial;

    float accumulationWeight = depthRejection;

    accumulationWeight *= normalWeight;
    accumulationWeight += clamp(((distance(history.rgb, current.rgb) / luma(history.rgb))) *
                                    (abs(cameraMovement.r) + abs(cameraMovement.g) + abs(cameraMovement.b)),
                                0, 1);
    accumulationWeight = clamp(1.0 - accumulationWeight, 0.0, 1);
    accumulationWeight = max(accumulationWeight, varianceSpatial);

    ///
    if (offscreen)
    {

        historyA = mix(0.0, clamp(historyA + (1 / (30)), 0, 1), float(distanceDelta < 0.2));

        if (hand)
            historyA = 1;
    }

    color.rgb = mix(history, current, clamp(accumulationWeight, 0, 0.9));

    color = clamp(color, 0.0001, 1000.0);
    vec4 currsspt = vec4(color, historyA);

    gl_FragData[0] = currsspt;
    gl_FragData[1] = vec4(clamp(moment, 0, 1), pack2(currentGData.rgb), clamp(currentGData.a / far, 0, 1));
}
