

#define waterfog
varying vec2 texcoord;

#include "/lib/fog/fog3.glsl"

float ld(const float dist)
{
    return (2.0 * near) / (far + near - dist * (far - near));
}
/////
vec3 median(sampler2D tex1, ivec2 coords)
{
    vec3 v[9];

    // Fill the array 'v' with the color values from the 3x3 pixel window
    for (int dX = -1; dX <= 1; ++dX)
    {
        for (int dY = -1; dY <= 1; ++dY)
        {
            ivec2 offset = ivec2(dX, dY);
            v[(dX + 1) * 3 + (dY + 1)] = texelFetch(tex1, coords + offset, 0).rgb;
        }
    }

    // Selection sort to find the 5th smallest element
    for (int i = 0; i < 5; ++i)
    {
        int minIndex = i;
        for (int j = i + 1; j < 9; ++j)
        {
            if (v[j].r + v[j].g + v[j].b < v[minIndex].r + v[minIndex].g + v[minIndex].b)
            {
                minIndex = j;
            }
        }

        // Swap to bring the current smallest to the i-th position
        vec3 temp = v[i];
        v[i] = v[minIndex];
        v[minIndex] = temp;
    }

    // Return the median value, which is the 5th smallest element
    return v[4];
}
/////
#ifdef DISTANT_HORIZONS
    #include "/lib/ssaodh.glsl"
#endif
vec3 getDepthPoint(vec2 coord, float depth)
{
    vec4 pos;
    pos.xy = coord;
    pos.z = depth;
    pos.w = 1.0;
    pos.xyz = pos.xyz * 2.0 - 1.0; // convert from the 0-1 range to the -1 to +1 range
    pos = gbufferProjectionInverse * pos;
    pos.xyz /= pos.w;

    return pos.xyz;
}
vec3 viewToScreenSpace(vec3 pos)
{
    vec3 screenPosition = projMAD(gbufferModelViewInverse, pos);
    screenPosition /= -pos.z;

    return screenPosition * 0.5 + 0.5;
}
vec3 waterNormal(vec2 coord)
{
    coord = mod(coord / 128.0, 1.0);
    float strength = 2.0;

    float deltaPos = 0.0015;

    vec2 offsets[4] = vec2[](vec2(deltaPos, 0.0), vec2(-deltaPos, 0.0), vec2(0.0, deltaPos), vec2(0.0, -deltaPos));

    float heights[4];
    for (int i = 0; i < 4; i++)
    {
        heights[i] = texture(colortex7, coord + offsets[i]).x;
    }

    float dx = (heights[1] - heights[0]) * strength;
    float dy = (heights[3] - heights[2]) * strength;
    vec3 full = normalize(vec3(dx, dy, 0.5)) + 0.5;

    return full;
}

vec2 computeRefractedCoord(vec2 texcoord, vec2 refractedCoord)
{

    vec4 normalTex = texture(colortex13, texcoord, 0);

    refractedCoord = normalTex.xy + texcoord;

    if (isEyeInWater == 0)
    {
        float z01 = ld(texture(depthtex0, refractedCoord).x);
        float z11 = ld(texture(depthtex1, refractedCoord).x);

        refractedCoord = mix(texcoord, refractedCoord, clamp(distance(z01, z11) * 512.0, 0, 1));
    }
    if (refractedCoord.x > 1.0 || refractedCoord.x < 0.0 || refractedCoord.y > 1.0 || refractedCoord.y < 0.0)
        refractedCoord = texcoord;

    return refractedCoord.xy;
}
vec2 computeRefractedCoordUnderwater(vec2 texcoord)
{

    float normalTex = texture(colortex7, texcoord * 0.5).x * 0.025;
    normalTex += texture(colortex7, texcoord * 0.75).x * 0.025;
    normalTex += texture(colortex7, texcoord * 0.25).x * 0.025;
    normalTex += texture(colortex7, texcoord * 0.1).x * 0.025;
    vec2 refractedCoord = normalTex + texcoord - 0.015;
    if (refractedCoord.x > 1.0 || refractedCoord.x < 0.0 || refractedCoord.y > 1.0 || refractedCoord.y < 0.0)
        refractedCoord = texcoord;
    return refractedCoord.xy;
}

vec3 GetWavesNormal(vec3 vector)
{
    return waterNormal(vector.xz - vector.y) * 2.0 - 1.0;
}

float CalculateWaterCaustics(vec3 screenSpacePosition)
{
    vec3 worldPos = mat3(gbufferModelViewInverse) * screenSpacePosition + cameraPosition;
    vec2 dither = blueNoise3(texcoord.st).xy;
    vec3 lightDir = -lightPos;

    float verticalDist = min(abs(worldPos.y - 63), 2.0);
    vec3 flatRefract = refract(lightDir, vec3(0.0, 1.0, 0.0), 1.0 / 1.3333);
    float rayLength = verticalDist / -flatRefract.y;
    vec3 center = worldPos - flatRefract * rayLength;

    const float threshold = 0.3;
    const int samples = 1;
    float caustics = 0.0;
    int count = 0;

    for (int i = -samples; i <= samples; i++)
    {
        for (int j = -samples; j <= samples; j++)
        {
            vec2 offset = vec2(i + dither.x, j + dither.y) * 0.15;
            vec3 lookup = center + vec3(offset.x, 0.0, offset.y);
            vec3 normal = GetWavesNormal(lookup).xzy;
            vec3 refractVec = refract(lightDir, normal, 1.0 / 1.3333);
            float len = verticalDist / refractVec.y;
            vec3 collision = lookup - refractVec * len;

            float distSq = dot(collision - worldPos, collision - worldPos) * 7.1;
            caustics += 1.0 - clamp(distSq / threshold, 0.0, 1.0);
            count++;
        }
    }

    caustics = (caustics / float(count)) / threshold;
    return pow(caustics, 1.0) * 2.0;
}

void main()
{
    /*RENDERTARGETS: 0,4*/

    vec2 refractedCoordTransp = texcoord;
    vec2 refractedCoord = texcoord;
    vec4 mask = texture(colortex5, refractedCoord);
    bool isWater1 = mask.r > 0.50;
    bool isWater = mask.r > 0.50;
    bool isGlassWater = mask.b > 0 && mask.b < 0.01;
    bool isGlass = mask.b > 0 && mask.b < 0.01;
    bool isGlassWater1 = mask.b > 0 && mask.b < 0.01;

    float z0 = texture(depthtex0, refractedCoord).x;
    float z1 = texture(depthtex1, refractedCoord).x;
    vec2 lightmap = vec2(0.0);
    bool isHand = z0 < 0.6;

#ifdef REFRACTION

    vec3 fragPos01 = toScreenSpace(vec3(refractedCoord, z0));
    vec3 fragPos11 = toScreenSpace(vec3(refractedCoord, z1));

    refractedCoord = computeRefractedCoord(texcoord, refractedCoord);
    refractedCoord = mix(refractedCoord, texcoord, GetBorderFogMixFactor(fragPos11, far, 0.0));

    refractedCoordTransp = mix(refractedCoordTransp, refractedCoord, float(isGlassWater || (isEyeInWater == 1 && isWater)));
    if (isEyeInWater == 1)
    {
        refractedCoord = computeRefractedCoordUnderwater(refractedCoord);
        refractedCoordTransp = refractedCoord;
    }
    z1 = texelFetch(depthtex1, ivec2(refractedCoord * viewResolution), 0).x;
    z0 = texelFetch(depthtex0, ivec2(refractedCoord * viewResolution), 0).x;
    mask = texture(colortex5, refractedCoord);
    if (isEyeInWater == 1)
    {
        isWater = mask.r > 0.25;
    }
    else
    {
        isWater = mask.r > 0.50;
    }
    isGlassWater = mask.b > 0 && mask.b < 0.01;
    isGlass = mask.b > 0 && mask.b < 0.01;

#endif
    vec4 cloud = vec4(0, 0, 0, 1.0);

#ifdef DISTANT_HORIZONS

    float z0DH = texture(dhDepthTex0, refractedCoord).x;
    float z1DH = texture(dhDepthTex1, refractedCoord).x;
    vec3 fragPos0 = toScreenSpace_alt(refractedCoord, z0, z0DH);
    vec3 fragPos1 = toScreenSpace_alt(refractedCoord, z1, z1DH);
    float borderFogFactor = GetBorderFogMixFactor(fragPos0, far, 0.0);

    z0 = (z0 >= 1.0) ? z0DH : z0;
    z1 = (z1 >= 1.0) ? z1DH : z1;
#else

    vec3 fragPos0 = toScreenSpace(vec3(refractedCoord, z0));
    vec3 fragPos1 = toScreenSpace(vec3(refractedCoord, z1));
    float borderFogFactor = GetBorderFogMixFactor(fragPos0, far, 0.0);

#endif

    vec3 color = texelFetch(colortex0, ivec2(refractedCoord * viewResolution), 0).rgb;

    lightmap = decodeVec2(texelFetch(colortex2, ivec2(refractedCoord * viewResolution), 0).b);
    vec3 lightCaustics;
#ifdef CAUSTICS
    if ((isEyeInWater == 0 && isWater) || (isEyeInWater == 1 && !isWater))
    {
        float caustics = CalculateWaterCaustics(fragPos1);

        lightCaustics = caustics * clamp(sunLight, 0.0, 1.0) * lightmap.y;
        lightCaustics = mix(vec3(1.0), lightCaustics, lightmap.y);
        color *= lightCaustics;
    }
#endif

    vec4 translucents = texelFetch(colortex1, ivec2(refractedCoordTransp * viewResolution), 0).rgba;
#ifdef WATER_FILTER
    if (isWater)
        translucents.rgb = median(colortex1, ivec2(refractedCoordTransp * viewResolution)).rgb;
#endif

    vec4 sky = texture(colortex8, refractedCoordTransp).rgba;
#ifdef CLOUDS
    #ifdef ALT_CLOUDS

    cloud = sky;

    if (z0 < 1.0 && !isHand && cloud.a < 1.0)
    {
        translucents.rgb *= clamp(cloud.a, 0, 1);
        translucents.rgb += cloud.rgb * clamp(1 - cloud.a, 0, 1);
    }
    if (z0 < 1.0 && !isHand && cloud.a < 1.0)
    {
        color.rgb *= clamp(cloud.a, 0, 1);
        color.rgb += cloud.rgb * clamp(1 - cloud.a, 0, 1);
    }

    #endif
#endif
    if (z0 < 1.0 || isEyeInWater > 0)

    {
        color = getVolumetricRays3(color.rgb, translucents, sky.rgb, borderFogFactor, fragPos0, fragPos1, isWater, clamp(lightmap.x + lightmap.y, 0.1, 1), clamp(lightmap.y * 2 - 0.5, 0, 1));
    }

    color *= exp(-length(fragPos0 / 2.0) * mix(darknessFactor, blindness, blindness));

    gl_FragData[0].rgb = vec3(color);

    float depthstuff = float(isWater1) * 0.5 + float(isGlassWater1) * 0.5;

#ifdef TAA
    gl_FragData[1].rgb = texture(colortex4, texcoord).rgb;
    gl_FragData[1].a = depthstuff;

#else
    gl_FragData[1].rgb = color;
    gl_FragData[1].a = depthstuff;
#endif
}
