
flat varying vec4 labvalue;
flat varying vec2 midtex;
varying vec4 lmtexcoord;
varying vec4 color;
flat varying float translucent;
flat varying float emissive;
varying vec3 normalMat;

#if defined NORMAL_MAPS || defined POM

varying vec4 normalMat2;
varying vec4 tangent;
    #if defined POM
        #include "/lib/pom.glsl"

    #endif
#endif

vec3 normalGenerator(vec2 coord, vec2 midtex)
{
    float noise = noise_standard(gl_FragCoord.xy);

    vec2 texSize = 1.0 / textureSize(texture, 0);

    vec2 deltaPos = texSize * noise * 4;

    vec2 offsets[4] = vec2[](vec2(deltaPos.x, 0.0), vec2(-deltaPos.x, 0.0), vec2(0.0, deltaPos.y), vec2(0.0, -deltaPos.y));
#ifdef LOD_FIX
    vec3 albedo2 = texture(texture, midtex).rgb;

    albedo2 += texture(texture, midtex + offsets[0]).rgb;
    albedo2 += texture(texture, midtex + offsets[1]).rgb;
    albedo2 += texture(texture, midtex + offsets[2]).rgb;
    albedo2 += texture(texture, midtex + offsets[3]).rgb;
    albedo2 += texture(texture, midtex + deltaPos).rgb;
    albedo2 += texture(texture, midtex - deltaPos).rgb;
    albedo2 /= 6;
#else
    vec3 albedo2 = texture2DLod(texture, coord, 4).rgb;
    albedo2 += texture(texture, midtex).rgb;
    albedo2 /= 2;
#endif

    float lumaLod = luma(albedo2);
    float heights[4];

    float strength = 1.0;
    vec2 texSize2 = texSize * 0.1 * noise;
    vec2 offsets1[4] = vec2[](vec2(texSize2.x, 0.0), vec2(-texSize2.x, 0.0), vec2(0.0, texSize2.y), vec2(0.0, -texSize2.y));

    for (int i = 0; i < 4; i++)
    {
        vec2 newcoord = coord + offsets1[i];
        if (distance(newcoord.x, midtex.x) > texSize.x * 8.0)
            newcoord.x = coord.x;
        if (distance(newcoord.y, midtex.y) > texSize.y * 8.0)
            newcoord.y = coord.y;
        float temp = luma(texture(texture, newcoord).rgb);

        heights[i] = 1 - (distance(temp, lumaLod));
    }

    float dx = (heights[1] - heights[0]) * strength;
    float dy = (heights[3] - heights[2]) * strength;
    vec3 full = normalize(vec3(dx, dy, 0.5)) + 0.5;
    full.z = 1.0;
    return full;
}

const int noiseSteps = 16;
const float noiseIntensity = 16.0;
const int noiseDropoff = 1024;

float rand(float co)
{
    return fract(sin(co * (91.3458)) * 47453.5453);
}
float rand(vec2 co)
{
    return fract(sin(dot(co.xy, vec2(12.9898, 78.233))) * 43758.5453);
}
float rand(vec3 co)
{
    return rand(co.xy + rand(co.z));
}

vec3 quantize(const in vec3 val, const in int stepSize)
{
    return floor(val * stepSize) / stepSize;
}
float toLinear2(float sRGB)
{
    return sRGB * (sRGB * (sRGB * 0.305306011 + 0.682171111) + 0.012522878);
}

void applyNoise(inout vec3 fragColor, const in vec3 worldPos, const in float viewDist)
{
    float noiseAmplification = noiseIntensity * 0.01;
    float lum = (fragColor.r + fragColor.g + fragColor.b) / 3.0;
    noiseAmplification =
        (1.0 - pow(lum * 2.0 - 1.0, 2.0)) * noiseAmplification; // Lessen the effect on depending on how dark the object
                                                                // is, equasion for this is -(2x-1)^{2}+1

    // Random value for each position
    float randomValue = rand(quantize(worldPos, noiseSteps));
    randomValue = randomValue * 1.5 * noiseAmplification - noiseAmplification;
    randomValue = (randomValue);

    fragColor.rgb = mix(vec3(randomValue), fragColor, 0.9);
}

// Constants for default values
const float DEFAULT_REFLECTANCE = 0.04;
const float SSS_MIN_THRESHOLD = 0.0;
const float SSS_MAX = 64.5 / 255.0;
const float EMISSIVE_MAX = 255.0;

// Helper function to calculate Subsurface Scattering (SSS)
float calculateSSS(float sssValue, vec4 labValue)
{
    if (sssValue == 0.0)
    {
        sssValue = labValue.z;
        if (sssValue > 0.0)
        {
            sssValue = (64.5 + ((sssValue * 255.0) * (255.0 - 64.5)) / 255.0) / 255.0;
        }
        else
        {
            sssValue = SSS_MIN_THRESHOLD;
        }
    }
    return sssValue;
}

// Helper function to calculate emissive properties
float calculateEmissive(float emissiveTexW, float emissive, vec3 albedoRGB)
{
    float emissiveValue = emissiveTexW * EMISSIVE_MAX < EMISSIVE_MAX ? emissiveTexW : 0.0;
    if (emissiveValue == 0.0)
    {
        emissiveValue = emissive * pow(luma(albedoRGB), 2.0);
    }
    return emissiveValue;
}

vec4 getSpecularProperties(vec2 tcoord, vec2 midtex, vec3 albedo, float translucent, float emissive)
{
    albedo = textureLod(texture, tcoord, 0.0).rgb;

    vec4 specularProperties;
    vec3 albedo2;
#ifndef ENTITY

    #ifdef LAB_PBR
    // LAB PBR Path
    vec4 specularTex = textureLod(specular, tcoord, 0.0);

    float smoothness = specularTex.x;
    float reflectance = specularTex.y > 0.0 ? specularTex.y : DEFAULT_REFLECTANCE;
    float sss = calculateSSS(specularTex.z, labvalue);

    float emissives = calculateEmissive(specularTex.w, emissive, albedo);

    #else
    // Non-LAB PBR Path
    float sss = labvalue.z;

    // Specific entity adjustments
    if (entityId == 1803)
    {
        sss = 1.0;
    }

    float emissives = emissive * pow(luma(albedo), 2.0);
    if (entityId == 1602)
    {
        emissives = 1.0;
    }

    // Noise and texture sampling
    float noiseVal = noise_standard(gl_FragCoord.xy);
    vec2 texSize = 1.0 / textureSize(texture, 0);
    vec2 deltaPos = texSize * noiseVal * 4.0;

    vec2 offsets[6] = vec2[](
        vec2(deltaPos.x, 0.0),
        vec2(-deltaPos.x, 0.0),
        vec2(0.0, deltaPos.y),
        vec2(0.0, -deltaPos.y),
        vec2(deltaPos),
        vec2(-deltaPos));

        #ifdef LOD_FIX
    // LOD Fix Path
    vec3 albedoAccum = texture(texture, midtex).rgb;
    for (int i = 0; i < 4; ++i)
    {
        albedoAccum += texture(texture, midtex + offsets[i]).rgb;
    }
    albedoAccum += texture(texture, midtex + offsets[4]).rgb;
    albedoAccum += texture(texture, midtex + offsets[5]).rgb;
    albedoAccum /= 6.0;
    albedo2 = albedoAccum;
        #else
    // Standard Path with LOD and Noise Application
    albedo2 = textureLod(texture, tcoord, 2.0).rgb;
            /*
                vec3 fragPos = toScreenSpace(gl_FragCoord.xyz * vec3(texelSize, 1.0));
                vec3 worldPos = mat3(gbufferModelViewInverse) * fragPos + gbufferModelViewInverse[3].xyz + cameraPosition;

                applyNoise(albedo2, worldPos + noiseVal * 0.01, far);

                */

        #endif

    // Calculate Luma and Modifier
    float lumaValue = luma(albedo);
    float lumaLod = luma(albedo2); // Assuming albedo2 represents a lower LOD
    float modifier = clamp(pow(abs(lumaValue - lumaLod), 0.25) + pow(abs(lumaValue - lumaLod), 0.5), 0.0, 1.0);

    // Reflectance and Smoothness Adjustment
    float reflectance = mix(labvalue.x * 1.2, labvalue.x * 0.7, modifier);
    float smoothness = mix(labvalue.y * 1.2, labvalue.y * 0.7, modifier);

    // Adjust SSS based on modifier
    if (sss > 0.0)
    {
        sss = (64.5 + ((mix(sss * 1.2, sss * 0.7, modifier) * 255.0)) * (255.0 - 64.5) / 255.0) / 255.0;
    }
    else
    {
        sss = (16.0 + ((mix(labvalue.w * 1.2, labvalue.w, modifier) * 255.0)) * (63.0 - 16.0) / 255.0) / 255.0;
    }

    // Final adjustment for high reflectance values
    if (labvalue.x >= 0.9)
    {
        reflectance = labvalue.x;
        smoothness = clamp(mix(labvalue.y * 1.2, labvalue.y * 0.7, modifier), 0.0, 1.0);
    }

    #endif
    specularProperties = vec4(sss, emissives, reflectance, smoothness);

#else
    specularProperties = vec4(0, 0, 0.02, 0.1);
#endif

    return specularProperties;
}

vec3 applyBump(mat3 tbnMatrix, vec3 bump)
{
    float bumpmultiplier = BUMP_MULT;
    float bumpmult = (1.0 - wetness * 0.95) * bumpmultiplier;

    bump = bump * vec3(bumpmult, bumpmult, bumpmult) + vec3(0.0f, 0.0f, 1.0f - bumpmult);

    return normalize(bump * tbnMatrix);
}

vec3 calculateNormal(vec2 tcoord, vec2 lmtexcoord, mat3 tbnMatrix, vec2 midtex, inout float ao)
{
    vec3 normalTex;
    vec3 normalout;

#ifdef GENERATED_NORMALS
    normalTex = normalGenerator(lmtexcoord.xy, midtex);
    normalout.xy = normalTex.xy * 2.0 - 0.996078431372549;
    normalout.z = sqrt(clamp(1.0 - dot(normalout.xy, normalout.xy), 0.0, 1.0));

#else
    normalTex = texture(normals, tcoord, 0).rgb;

    ao *= normalTex.b;

    normalout.xy = normalTex.xy * 2.0 - 0.996078431372549;
    normalout.z = sqrt(clamp(1.0 - dot(normalout.xy, normalout.xy), 0.0, 1.0));
#endif

    normalout = applyBump(tbnMatrix, normalout);

    return normalout;
}
float lmfaloff(float lm)
{
    float falloff = 2.0;

    lm = exp(-(1.0 - lm) * falloff);
    lm = max(0.0, lm - exp(-falloff));

    return lm * 2;
}
mat2 mat2_rotate_z(float radians)
{
    return mat2(
        cos(radians), -sin(radians),
        sin(radians), cos(radians));
}

const vec3[] COLORS = vec3[](
    vec3(0.022087, 0.098399, 0.110818),
    vec3(0.011892, 0.095924, 0.089485),
    vec3(0.027636, 0.101689, 0.100326),
    vec3(0.046564, 0.109883, 0.114838),
    vec3(0.064901, 0.117696, 0.097189),
    vec3(0.063761, 0.086895, 0.123646),
    vec3(0.084817, 0.111994, 0.166380),
    vec3(0.097489, 0.154120, 0.091064),
    vec3(0.106152, 0.131144, 0.195191),
    vec3(0.097721, 0.110188, 0.187229),
    vec3(0.133516, 0.138278, 0.148582),
    vec3(0.070006, 0.243332, 0.235792),
    vec3(0.196766, 0.142899, 0.214696),
    vec3(0.047281, 0.315338, 0.321970),
    vec3(0.204675, 0.390010, 0.302066),
    vec3(0.080955, 0.314821, 0.661491));

const mat4 SCALE_TRANSLATE = mat4(
    0.5, 0.0, 0.0, 0.25,
    0.0, 0.5, 0.0, 0.25,
    0.0, 0.0, 1.0, 0.0,
    0.0, 0.0, 0.0, 1.0);

mat4 end_portal_layer(float layer)
{
    mat4 translate = mat4(
        1.0, 0.0, 0.0, 17.0 / layer,
        0.0, 1.0, 0.0, (2.0 + layer / 1.5) * (frameTimeCounter * 0.01),
        0.0, 0.0, 1.0, 0.0,
        0.0, 0.0, 0.0, 1.0);

    mat2 rotate = mat2_rotate_z(radians((layer * layer * 4321.0 + layer * 9.0) * 4.0));

    mat2 scale = mat2((4.5 - layer / 4.0) * 2.0);

    return mat4(scale * rotate) * translate * SCALE_TRANSLATE;
}
float interleaved_gradientNoise()
{
    return fract(52.9829189 * fract(0.06711056 * gl_FragCoord.x + 0.00583715 * gl_FragCoord.y) + frameTimeCounter * 51.9521);
}
#if defined POM
vec2 applyPOM(vec2 tcoord, vec3 fragpos)
{
    // Compute TBN matrix
    vec3 tangent2 = normalize(cross(tangent.rgb, normalMat2.rgb) * tangent.w);
    mat3 tbnMatrix2 = mat3(
        tangent.x, tangent2.x, normalMat2.x,
        tangent.y, tangent2.y, normalMat2.y,
        tangent.z, tangent2.z, normalMat2.z);

    // Apply POM
    vec2 newTcoord = pomColor(tcoord, fragpos, interleaved_gradientNoise(), tbnMatrix2);
    if (readNormal(lmtexcoord.st).a == 1)
        newTcoord = tcoord;

    return newTcoord;
}
#endif
vec4 getMaterialProperties(vec2 tcoord, vec3 albedoRGB)
{
    return getSpecularProperties(tcoord, midtex, albedoRGB, translucent, emissive);
}
vec3 calculateNormalWrapper(vec2 tcoord, float ao, vec3 fragpos)
{

    vec3 normal;
#ifdef NORMAL_MAPS

    vec3 tangent2 = normalize(cross(tangent.rgb, normalMat2.rgb) * tangent.w);
    mat3 tbnMatrix2 = mat3(
        tangent.x, tangent2.x, normalMat2.x,
        tangent.y, tangent2.y, normalMat2.y,
        tangent.z, tangent2.z, normalMat2.z);

    normal = viewToWorld(calculateNormal(tcoord, lmtexcoord.xy, tbnMatrix2, midtex, ao));
#else
    normal = normalMat;
#endif

#ifdef NORMAL_FIX
    vec3 cameraVector = normalize(fragpos);
    vec3 vnormal = worldToView(normalMat);
    normal = viewToWorld(faceforward(worldToView(normal), cameraVector, vnormal));
#endif

    return normal;
}
vec2 computeAdvancedLighting(vec2 lightmap, vec3 fragpos, vec3 normal)
{
#ifdef DLM
    // Compute derivatives
    vec3 dx = dFdx(fragpos);
    vec3 dy = dFdy(fragpos);
    vec3 dz = cross(dx, dy) * 0.01;

    vec2 derivation = 256.0 * vec2(dFdx(lightmap.x), dFdy(lightmap.x));
    vec3 direction = normalize(derivation.x * dx + derivation.y * dy + dz);

    float ndotd = dot(worldToView(normal), direction);
    float power = 1.0 - ndotd * lightmap.x;
    float derivationLength = min(length(derivation) * 10.0, 1.0);

    lightmap.x = mix(lightmap.x, pow(lightmap.x, power), derivationLength);
#endif

    return lightmap;
}
vec3 applyPBRAdjustments(vec3 albedoRGB, vec4 materialProps, vec2 lightmap)
{
    albedoRGB = toLinear(albedoRGB);

    // Adjust materials based on PBR mode
    vec2 materials1 = vec2(color.a, materialProps.x);
    lightmap.x += materialProps.y;

    if (entityId == 1602)
    {
        materials1.r = 1.0;
    }

    albedoRGB += (materialProps.y * 8.0 * albedoRGB * albedoRGB);
    lightmap = min(lightmap, 1.0);

    return albedoRGB;
}
void outputToFramebuffer(vec4 albedo, vec3 normal, vec2 lightmap, vec4 materialProps, float ao)
{

    gl_FragData[0] = albedo; // Albedo
    gl_FragData[1].rgba = vec4(
        encodeNormal(normal),
        encodeVec2(lightmap),
        encodeVec2(vec2(ao, materialProps.x))); // Normals and materials
    gl_FragData[2].r = encodeVec2(materialProps.zw);
#ifdef ENTITY
    gl_FragData[3].g = 1.0;
#endif
}
#ifdef ENTITY
/* RENDERTARGETS: 0,2,3,5 */
#else
/* RENDERTARGETS: 0,2,3 */
#endif
void main()
{
    vec2 tcoord = lmtexcoord.xy;
    vec2 lightmap = lmtexcoord.zw;

    // Compute fragment position in screen space
    vec3 fragpos = toScreenSpace(gl_FragCoord.xyz * vec3(texelSize, 1.0));

// Parallax Occlusion Mapping (POM)
#if defined POM
    tcoord = applyPOM(tcoord, fragpos);
#endif

    // Fetch textures and material properties
    vec4 albedo = texture(texture, tcoord, 0);
    vec4 materialProps = getMaterialProperties(tcoord, albedo.rgb);

    // Apply color modifications
    albedo.rgb *= color.rgb;
#ifdef ENTITY
    if (entityId == 1600) // not a good fix, will sort this later.tm
    {
        discard;
    }

    albedo.rgb = mix(albedo.rgb, entityColor.rgb, entityColor.a);
#endif

    float ao = color.a;

    // Calculate normals
    vec3 normal = calculateNormalWrapper(tcoord, ao, fragpos);

// Additional lighting computations
#ifdef PBR
    lightmap = computeAdvancedLighting(lightmap, fragpos, normal);
#endif

    // Apply PBR adjustments
    albedo.rgb = applyPBRAdjustments(albedo.rgb, materialProps, lightmap);

    // Output to framebuffer
    outputToFramebuffer(albedo, normal, lightmap, materialProps, color.a);
}
