flat varying vec2 midtex;
flat varying vec4 labvalue;
varying vec4 lmtexcoord;
varying vec4 color;
flat varying float emissive;

attribute vec4 mc_Entity;
attribute vec4 mc_midTexCoord;
attribute vec4 at_tangent;

#if defined NORMAL_MAPS || defined POM

varying vec4 vtexcoordam;
varying vec4 vtexcoord;
varying vec4 tangent;
varying vec4 normalMat2;

#endif
varying vec3 normalMat;

#define diagonal3(m) vec3((m)[0].x, (m)[1].y, m[2].z)
#define projMAD(m, v) (diagonal3(m) * (v) + (m)[3].xyz)

#include "/lib/vertex/main.glsl"

// Helper function to compute emissive value based on entity check
float computeEmissive(float lmtexcoordZ, int entityCheck)
{
    if (entityCheck >= 100 && entityCheck <= 355)
        return pow(lmtexcoordZ, 8.0);
    return 0.0;
}

// Helper function to compute TBN Matrix
#if defined NORMAL_MAPS || defined POM
mat3 computeTBN(vec3 normal)
{
    return computeTBNMatrix(gbufferModelViewInverse, gl_NormalMatrix, at_tangent, gl_Normal);
}
#endif

// Helper function to compute the final position
vec3 computePosition(vec4 vertex)
{
    return (mat3(gl_ModelViewMatrix) * vertex.xyz) + gl_ModelViewMatrix[3].xyz;
}

// Main function
void main()
{
    lmtexcoord.xyzw = vec4(gl_MultiTexCoord0.xy, gl_MultiTexCoord1.xy * recip);
    emissive = 0.0;
    labvalue = vec4(0);

    int entityCheck = int(mc_Entity.x);
    labvalue = getMaterialProperties(entityCheck);

    emissive = computeEmissive(lmtexcoord.z, entityCheck);
    color = gl_Color;
    midtex.xy = mc_midTexCoord.st;
#if defined NORMAL_MAPS || defined POM || defined COMPLEX_NORMALS
    vec3 normalMat0 = normalize(gl_NormalMatrix * gl_Normal);
#endif
#if defined COMPLEX_NORMALS || defined POM
    normalMat = viewToWorld(normalMat0);

#else
    normalMat = gl_Normal;
#endif

#if defined NORMAL_MAPS || defined POM
    vec2 midcoord = (gl_TextureMatrix[0] * mc_midTexCoord).st;
    normalMat2 = vec4(normalMat0, 0);
    bool tangentValid = dot(at_tangent.xyz, at_tangent.xyz) > 1e-9;
    vec3 T = tangentValid ? normalize(mat3(gl_NormalMatrix) * at_tangent.xyz) : vec3(0.0);
    tangent = vec4(normalize(T), at_tangent.w);
    vec2 texcoordminusmid = lmtexcoord.xy - midcoord;
    vtexcoordam.pq = abs(texcoordminusmid) * 2.0;
    vtexcoordam.st = min(lmtexcoord.xy, midcoord - texcoordminusmid);
    vtexcoord.xy = sign(texcoordminusmid) * 0.5 + 0.5;
#endif

    vec3 position = computePosition(gl_Vertex);

#ifdef LEAVES
    bool Flower = isFlower(entityCheck, (gl_MultiTexCoord0.t < mc_midTexCoord.t));
    bool Leaves = isLeaves(entityCheck, (abs(position.z) < 128.0 && gl_MultiTexCoord1.y > 100.0));
    if (Flower || Leaves)
    {
        position = computeWaveOffset(position, cameraPosition, gl_Vertex, gl_MultiTexCoord1, gbufferModelView);
    }
#endif

#if defined ENTITY && defined NO_ENTITY_TAA
    gl_Position = toClipSpace4(position);
#else
    gl_Position = taaFunc(toClipSpace4(position));
#endif
}
