#extension GL_ARB_gpu_shader5 : enable

varying vec2 texcoord;

vec2 HashVec(vec2 d)
{
    float a = dot(d, vec2(123.4, 234.5));
    float b = dot(d, vec2(234.5, 345.6));
    return sin(58322.51 * sin(vec2(a, b)) + frameTimeCounter * 2.0);
}

vec2 wrapCoordinates(vec2 coords, float size)
{
    return mod(coords, size); // Ensure coordinates fall within the grid
}
float map(float value, float min1, float max1, float min2, float max2)
{
    return min2 + (value - min1) * (max2 - min2) / (max1 - min1);
}
// Generates 1 Octave of Perlin Noise
float Octave(vec2 uv, float GridSize)
{
    vec2 GridUv = fract(uv * GridSize);
    vec2 CellId = floor(uv * GridSize);
    vec2 SmoothUv = smoothstep(0.0, 1.0, GridUv);

    float n00 = dot(HashVec(wrapCoordinates(CellId, GridSize)), GridUv);
    float n10 = dot(HashVec(wrapCoordinates(CellId + vec2(1.0, 0.0), GridSize)), GridUv - vec2(1.0, 0.0));
    float n01 = dot(HashVec(wrapCoordinates(CellId + vec2(0.0, 1.0), GridSize)), GridUv - vec2(0.0, 1.0));
    float n11 = dot(HashVec(wrapCoordinates(CellId + vec2(1.0, 1.0), GridSize)), GridUv - vec2(1.0, 1.0));

    float nx0 = mix(n00, n10, SmoothUv.x);
    float nx1 = mix(n01, n11, SmoothUv.x);
    return mix(nx0, nx1, SmoothUv.y);
}

float WaterMap(vec2 pos)
{
    float gradient = 0.0;
    float frequency = 8.0;
    float amplitude = 6.0;

    for (int i = 0; i < 6; i++)
    {
        gradient += (Octave(pos, frequency) + 0.5) / amplitude;
        frequency *= 3.0;
        amplitude *= 1.75;
    }

    return gradient * 0.75;
}
float CausticsMap(vec2 pos)
{
    vec2 e = vec2(0.1, 0.0);

    // Compute the heightmap at the current position
    float h = WaterMap(pos);

    // Compute the heightmap at neighboring positions to calculate the gradient
    float hx = WaterMap(pos + e.xy);
    float hy = WaterMap(pos + e.yx);
    vec2 grad = vec2(hx - h, hy - h) / e.x;

    // Compute the length of the gradient vector
    float l = length(grad);

    // Calculate the caustics intensity using an inverse square of the gradient length
    float intensity = 0.5 / (l * l + 0.1);

    // Optionally, adjust the intensity range to suit your needs
    return clamp(pow(intensity, 0.5) * 0.25, 0.0, 1.0);
}

void main()
{
    /* DRAWBUFFERS:7*/
    vec3 fragpos = toScreenSpace(vec3(texcoord, 1));

    float water = WaterMap(texcoord);
    float casutic = CausticsMap(texcoord);

    gl_FragData[0].r = water;
    gl_FragData[0].g = casutic;
}
